    gSyntaxTests = pa_fetchSyntax()
    
    gpa_ErrorMessage = ""
    gpa_ErrorLine = -1         // inside this library scripts start at line 0

// -----------------------------  parsing ---------------------------

    // At this point the script should be parsed.  The parsing routine should
    // accept one parameter (the script as text) and return an object
    // containing int, text, array, array
    // [errorLine, errorMessage, script blocks, lineRefs]

function pa_parseScript(theScript) {
    gpa_ErrorMessage = ""
    gpa_ErrorLine = -1
    
    var lineRefs = []
    var scriptLines = theScript.split(gReturn)
    var scriptLineCount = scriptLines.length

    // get the various blocks
    var blocks = pa_getBlocks(scriptLines, scriptLineCount)
    if (gpa_ErrorLine == -1) {
        var lineRefs = pa_composeLineRefs(blocks, scriptLineCount)
    }
    
    return {errorLine: gpa_ErrorLine,
            errorMessage: gpa_ErrorMessage,
            blocks: blocks,
            lineRefs: lineRefs}
}

function pa_composeLineRefs(blocks, scriptLineCount) {
    var lineRefs = []
    
    var namedThings = []
    
    // prescan for named things like tables and procs and spot duplicates
    for (var i=0; i<blocks.length; i++) {
        var thisBlock = blocks[i]
        switch (thisBlock.blockType) {
            case 'table':
            case 'proc':
            case 'textblock':
            case 'defineline':
            case 'comment':
                var temp = thisBlock.blockName
                if (thisBlock.blockType == 'table') {
                    if (temp.indexOf(" in random order") != -1) {
                        temp = temp.substring(0,temp.length-16)
                    }
                }
                
                var searchThing = thisBlock.blockType+' '+temp
                var temp = util_lookInObjList(namedThings, searchThing, -1)
                if (temp == -1) {   // not there yet
                
                    namedThings.push({theKey: searchThing,
                                        startLine: thisBlock.startLine})
                                        
                } else {            // already got that thing
                    gpa_ErrorLine = thisBlock.startLine
                    gpa_ErrorMessage = "Found more than one '"+searchThing+"'"
                    return lineRefs
                }
                break
        }
    }
    
    // check to see that there's a 'proc main'
    var temp = util_lookInObjList(namedThings, 'proc main', -1)
    if (temp == -1) {
        gpa_ErrorLine = scriptLineCount-1
        gpa_ErrorMessage = "Can't find a 'proc main'"
        return lineRefs
    }
    
    // check to the 'define fileFormat'
    var temp = util_lookInObjList(namedThings, 'defineline fileFormat', -1)
    if (temp == -1) {
        gpa_ErrorLine = scriptLineCount-1
        gpa_ErrorMessage = "Can't find a fileFormat definition"
        return lineRefs
    }
    
    // check to see that there's a 'define homeFolder'
    var temp = util_lookInObjList(namedThings, 'defineline homeFolder', -1)
    if (temp == -1) {
        gpa_ErrorLine = scriptLineCount-1
        gpa_ErrorMessage = "Can't find a homeFolder definition"
        return lineRefs
    }
    
    // check to see that there's a 'define homeFolder'
    var temp = util_lookInObjList(namedThings, 'defineline logMethod', -1)
    if (temp == -1) {
        gpa_ErrorLine = scriptLineCount-1
        gpa_ErrorMessage = "Can't find a logMethod definition"
        return lineRefs
    }
    
    // start off lineRefs as empty strings
    // might better do this by cloning the beginning of an array of empty
    // strings
    for (var i=scriptLineCount-1; i>=0; i--) {
        lineRefs.push("X")
    }
    
    // now every line is an "X", try and get rid of all the "X" lines
    for (var i=0; i<blocks.length; i++) {
        var thisBlock = blocks[i]
        var thisStart = thisBlock.startLine
        
        switch (thisBlock.blockType) {
            case 'blankline':
                lineRefs[thisStart] = ''
                break
            
            case 'commentline':
                lineRefs[thisStart] = '--'
                break
            
            case 'defineline':
                lineRefs[thisStart] = 'def'
                break
            
            case 'proc':
                for (var j=0; j<thisBlock.linesAsParams.length; j++) {
                    var paramType = thisBlock.linesAsParams[j][0]
                    if (paramType == 'notUnderstood') {
                        lineRefs[thisStart+j] = 'ERROR'
                        gpa_ErrorLine = thisBlock.startLine+j
                        gpa_ErrorMessage = "Don't understand script line '"+thisBlock.linesAsParams[j][1]+"'"
                        return lineRefs
                    } else if (paramType == 'blankLine') {
                        lineRefs[thisStart+j] = ' '
                    } else if (paramType == 'halt') {
                        lineRefs[thisStart+j] = ' halt'
                    } else if (paramType == 'commentLine') {
                        lineRefs[thisStart+j] = ' -'
                    } else if (paramType == 'procEnd') {
                        lineRefs[thisStart+j] = 'pe'
                    } else if (paramType == 'proc') {
                        lineRefs[thisStart+j] = util_replaceAll(lineRefs[thisStart+j], 'X', 'ps')
                    } else if (paramType == 'call') {
                    
                        lineRefs[thisStart+j] = 'call'
                        var searchThing = 'proc '+thisBlock.linesAsParams[j][1]
                        var searchLine = util_lookInObjList(namedThings, searchThing, {startLine: -1}).startLine
                        if (searchLine == -1) {     // no proc for call
                        
                            lineRefs[thisStart+j] = 'ERROR'
                            gpa_ErrorLine = thisBlock.startLine+j
                            gpa_ErrorMessage = "Call to non-existant "+searchThing
                            return lineRefs
                        
                        } else if (searchLine > thisStart+j) {
                        
                            lineRefs[thisStart+j] = " v"
                            lineRefs[searchLine] += "v"
                        
                        } else {
                        
                            lineRefs[thisStart+j] = " ^"
                            lineRefs[searchLine] += "^"
                        
                        }
                    } else if (paramType == 'readFromTable') {
                        
                        lineRefs[thisStart+j] = 'read'
                        var searchThing = 'table '+thisBlock.linesAsParams[j][2]
                        var searchLine = util_lookInObjList(namedThings, searchThing, {startLine: -1}).startLine
                        if (searchLine == -1) {     // no table for read
                            
                            lineRefs[thisStart+j] = 'ERROR'
                            gpa_ErrorLine = thisBlock.startLine+j
                            gpa_ErrorMessage = "Read from non-existant table "+searchThing
                            return lineRefs
                            
                        } else if (searchLine > thisStart+j) {
                            
                            lineRefs[thisStart+j] = " v"
                            lineRefs[searchLine] += "v"
                            
                        } else {
                            
                            lineRefs[thisStart+j] = " ^"
                            lineRefs[searchLine] += "^"
                            
                        }
                        
                    } else if (paramType == 'displaySurvey') {
                        
                        lineRefs[thisStart+j] = 'disp'
                        var searchThing = 'table '+thisBlock.linesAsParams[j][1]
                        var searchLine = util_lookInObjList(namedThings, searchThing, {startLine: -1}).startLine
                        if (searchLine == -1) {     // no table for displaySurvey
                            
                            lineRefs[thisStart+j] = 'ERROR'
                            gpa_ErrorLine = thisBlock.startLine+j
                            gpa_ErrorMessage = "Display survey from non-existant table "+searchThing
                            return lineRefs
                            
                        } else if (searchLine > thisStart+j) {
                            
                            lineRefs[thisStart+j] = " v"
                            lineRefs[searchLine] += "v"
                            
                        } else {
                            
                            lineRefs[thisStart+j] = " ^"
                            lineRefs[searchLine] += "^"
                            
                        }
                        
                    } else if (paramType == 'displayText') {
                    
                        lineRefs[thisStart+j] = 'display'
                        var searchThing = 'textblock '+thisBlock.linesAsParams[j][1]
                        if (searchThing.substr(0,1) != '$') {
                            var searchLine = util_lookInObjList(namedThings, searchThing, {startLine: -1}).startLine
                            if (searchLine == -1) {     // no proc for call
                            
                                lineRefs[thisStart+j] = 'ERROR'
                                gpa_ErrorLine = thisBlock.startLine+j
                                gpa_ErrorMessage = "displaying non-existant "+searchThing
                                return lineRefs
                           
                            } else if (searchLine > thisStart+j) {
                            
                                lineRefs[thisStart+j] = " v"
                                lineRefs[searchLine] += "v"
                            
                            } else {
                            
                                lineRefs[thisStart+j] = " ^"
                                lineRefs[searchLine] += "^"
                            
                            }
                        }
                
                    } else {
                        lineRefs[thisStart+j] = ' .'
                    }
                }
                break
            
            case 'table':
                lineRefs[thisBlock.startLine] = util_replaceAll(lineRefs[thisBlock.startLine], 'X', 'ts')
                for (var j=0; j<thisBlock.content.length; j++) {
                    lineRefs[thisStart+1+j] = ' r'
                }
                lineRefs[thisStart+j+1] = 'te'
                break
                
            case 'textblock':
                lineRefs[thisBlock.startLine] = 'ts'
                for (var j=0; j<thisBlock.content.length; j++) {
                    lineRefs[thisStart+1+j] = ' t'
                }
                lineRefs[thisStart+j+1] = 'te'
                break
                
            case 'commentblock':
                lineRefs[thisBlock.startLine] = 'cs'
                for (var j=0; j<thisBlock.content.length; j++) {
                    lineRefs[thisStart+1+j] = ' c'
                }
                lineRefs[thisStart+j+1] = 'ce'
                break
                
            default:
                alert("Parser doesn't recognise block type "+thisBlock.blockType+".")
        }
    }

    // check to see that {if, while, repeat} nest properly.
    var nestStack = []
    for (var i=0; i<blocks.length; i++) {
        var thisBlock = blocks[i]
        if (thisBlock.blockType == 'proc') {
        
            var thisStart = thisBlock.startLine
            for (var j=0; j<thisBlock.linesAsParams.length; j++) {
                var theseStacks = util_lookInObjList(gStepTypeNesting, thisBlock.linesAsParams[j][0], {unstack:"", stack:""})
                if (theseStacks.unstack != "") {
                
                    if (nestStack.length == 0) {
                        lineRefs[thisStart+j] = 'ERROR'
                        gpa_ErrorLine = thisBlock.startLine+j
                        gpa_ErrorMessage = "Expected to be in a '"+theseStacks.unstack+"' but wasn't"
                        return lineRefs
                    } else {
                        var topOfStack = nestStack.pop()
                        if (topOfStack != theseStacks.unstack) {
                            lineRefs[thisStart+j] = 'ERROR'
                            gpa_ErrorLine = thisBlock.startLine+j
                            gpa_ErrorMessage = "Expected to be in a '"+theseStacks.unstack+"' but was in a '"+topOfStack+"'"
                        return lineRefs
                        }
                    }
                
                }
                if (theseStacks.stack != "") {
                    nestStack.push(theseStacks.stack)
                }
            }
            if (nestStack.length != 0) {
                lineRefs[thisStart+j] = 'ERROR'
                gpa_ErrorLine = thisBlock.startLine+j
                gpa_ErrorMessage = "Reached end of procedure while inside a '"+nestStack.pop()+"'"
            }
        }
    }

    // tidy up lineRefs for display
    for (var i=scriptLineCount-1; i>=0; i--) {
        if (lineRefs[i].length > 6) {
            lineRefs[i] = lineRefs[i].substr(0,6)
        }
    }
    
    return lineRefs
}

// internal structure of blocks is a four-item array:
//
// [0] is the number of the line it starts on
// [1] is the type of the item as text: 'proc', 'table', 'textblock', 'comment'
// [2] is the name of the item 'main' or the text of the comment for comment
// [3] is the content as simple text
// [4] is the content structured, or for comment doesn't exist

function pa_getBlocks(scriptLines) {
    var legitBlockTypes = ["comment", "proc", "table", "textblock"]
    var legitBlockEnds = ["end comment", "end proc", "end table", "end textblock"]

    var blockArray = []
    
    var currentBlockType = "none"
    var currentBlockName = "none"
    var currentBlockStartLine = -1
    var expecting = "not initialised"

    for (var currentLineNum=0; currentLineNum<scriptLines.length; currentLineNum++) {
        var thisLine = scriptLines[currentLineNum].trim()
        var firstSpacePos = thisLine.indexOf(gSpace)
        if (firstSpacePos == -1) {
            var theFirstWord = thisLine
            var theRest = ""
        } else {
            var theFirstWord = thisLine.substring(0,firstSpacePos)
            var theRest = thisLine.substring(firstSpacePos+1)
        }
        
        if (currentBlockType == "none") {
        
            if (thisLine == "") {
            
                var asStruct = pa_parseTextToBlankline(
                        currentLineNum,
                        theRest,
                        theRest)
                blockArray.push(asStruct)
                        
            } else if (theFirstWord == "--") {
            
                var asStruct = pa_parseTextToCommentline(
                        currentLineNum,
                        theRest,
                        theRest)
                blockArray.push(asStruct)
                        
            } else if (theFirstWord == "define") {
            
                var spacePos = theRest.indexOf(" ")
                if (spacePos == -1) {
                    gpa_ErrorLine = currentLineNum
                    gpa_ErrorMessage = "Found a definition line with no space"
                    currentLineNum = scriptLines.length
                }
                var asStruct = pa_parseTextToDefineline(
                        currentLineNum,
                        theRest.substr(0,spacePos),
                        theRest.substr(spacePos+1))
                blockArray.push(asStruct)
                        
            } else if (legitBlockTypes.indexOf(theFirstWord) == -1) {
            
                gpa_ErrorLine = currentLineNum
                gpa_ErrorMessage = "Found '"+thisLine+"' instead of a block or the end of the script"
                currentLineNum = scriptLines.length

            } else {        // it's a legitimate start of a block !
            
                var currentBlockType = theFirstWord
                var currentBlockName = theRest
                var currentBlockStartLine = currentLineNum
                var expecting = "end "+ currentBlockType
                
            }
            
        } else {                    // we're in a block
        
            if (thisLine == expecting) {        // reached end of current block
            
                // handle the block
                switch(currentBlockType) {
                    case 'comment':
                        var asStruct = pa_parseTextToCommentblock(
                                currentBlockStartLine,
                                currentBlockName,
                                scriptLines.slice(currentBlockStartLine, currentLineNum+1))
                        blockArray.push(asStruct)
                        break
                    case 'proc':
                        var asStruct = pa_parseTextToProc(
                                currentBlockStartLine,
                                currentBlockName,
                                scriptLines.slice(currentBlockStartLine, currentLineNum+1))
                        blockArray.push(asStruct)
                        break
                    case 'table':
                        var asStruct = pa_parseTextToTable(
                                currentBlockStartLine,
                                currentBlockName,
                                scriptLines.slice(currentBlockStartLine, currentLineNum+1))
                        blockArray.push(asStruct)
                        break
                    case 'textblock':
                        var asStruct = pa_parseTextToTextblock(
                                currentBlockStartLine,
                                currentBlockName,
                                scriptLines.slice(currentBlockStartLine, currentLineNum+1))
                        blockArray.push(asStruct)
                        break
                }
                
                // no lonoger in block, expect a new block
                var currentBlockType = "none"
                var currentBlockName = "none"
                var currentBlockStartLine = -1
            
            } else if (legitBlockEnds.indexOf(theFirstWord) != -1) {
            
                // trying to end the wrong type of block
                gpa_ErrorLine = currentLineNum
                gpa_ErrorMessage = "Found '"+thisLine+"' when I expected '"+expecting+"'"
                currentLineNum = scriptLines.length
                
            } else if (legitBlockTypes.indexOf(theFirstWord) != -1) {
            
                // trying to start a new block while already in one
                gpa_ErrorLine = currentLineNum
                gpa_ErrorMessage = "Found '"+thisLine+"' when I was already in a block of type '"+currentBlockType+"'"
                currentLineNum = scriptLines.length
            }
        }
    }
        
    // got to the end of the script
    if ((gpa_ErrorLine == -1) && (currentBlockType != "none")) {
        gpa_ErrorLine = currentLineNum
        gpa_ErrorMessage = "Expected '"+expecting+"' but reached the end of the script"
    }
    
    return blockArray
}

// -------------------- parsers for each type ------------------------

function pa_parseTextToBlankline(startLineNum, blockName, raw) {
    return {startLine: startLineNum,
            blockType: "blankline",
            blockName: blockName,
            content: raw}
}

function pa_parseTextToCommentline(startLineNum, blockName, raw) {
    return {startLine: startLineNum,
            blockType: "commentline",
            blockName: blockName,
            content: raw}
}

function pa_parseTextToDefineline(startLineNum, blockName, raw) {
    return {startLine: startLineNum,
            blockType: "defineline",
            blockName: blockName,
            content: raw}
}

function pa_parseTextToCommentblock(startLineNum, blockName, raw) {
    var soFar = raw.slice(1,-1)
    
    return {startLine: startLineNum,
            blockType: "commentblock",
            blockName: blockName,
            content: soFar}
}

function pa_parseTextToTextblock(startLineNum, blockName, raw) {
    var soFar = raw.slice(1,-1)
    
    return {startLine: startLineNum,
            blockType: "textblock",
            blockName: blockName,
            content: soFar}
}

function pa_parseTextToTable(startLineNum, blockName, raw) {
    var strucSoFar = []
    var arraySoFar = []
    
    for (var i=1; i<raw.length-1; i++) {
        thisLine = util_replaceAll(raw[i].trim(), gComma, gTab)
        
        strucSoFar.push(thisLine.split(gTab))
        if (thisLine != "") arraySoFar.push(thisLine.split(gTab))
    }
    
    return {startLine: startLineNum,
            blockType: "table",
            blockName: blockName,
            content: strucSoFar,
            rowsAsArray: arraySoFar}
}

// ----------------------- parsing proc --------------------------

function pa_parseTextToProc(startLineNum, blockName, raw) {
    var paramsSoFar = []
    var textSoFar = []
    
    for (var i=0; i<raw.length; i++) {
        var thisLineAsText = raw[i].trim()
        textSoFar.push(thisLineAsText)
        
        var thisLineAsParams = pa_identifyStep(thisLineAsText)
        paramsSoFar.push(thisLineAsParams)
    }
    
    return {startLine: startLineNum,
            blockType: "proc",
            blockName: blockName,
            linesAsParams: paramsSoFar,
            linesAsText: textSoFar}
}

// try to identify which entry in gStepTypes this line corresponds to theText
// if can't, then return it as if it's a "notUnderstood" step
function pa_identifyStep(theText) {
    for (var i = gSyntaxTests.length-1; i>=0; i--) {
        var temp = pa_listVariableParts(gSyntaxTests[i], theText)
        if (temp != -1) return temp
    }
    
    return ["notUnderstood", theText]
}

// if a match returns [stepType, p1, p2, p3 …] else returns -1
function pa_listVariableParts(theFixeds, theText) {
    var textSoFar = theText
    var varsSoFar = [theFixeds[0]]
    var fixedsSoFar = theFixeds.slice(1)
    
    if (fixedsSoFar.length == 1) {      // we have one piece of text left
        if (textSoFar == fixedsSoFar[0]) {
            return varsSoFar
        } else {
            return -1
        }
    } else {    // we have first fixed text and then a variable to strip
        // first get rid of the fixed bit at the end
        var thisFixed = fixedsSoFar.pop()
        var theLength = thisFixed.length
        if (thisFixed != "") {
            var lastPart = textSoFar.substring(textSoFar.length-thisFixed.length)
            if (lastPart != thisFixed) {
                return -1
            } else {
                textSoFar = textSoFar.substring(0,textSoFar.length-thisFixed.length)
            }
        }
        
        // now we have left F v F v F v so peel off the beginning
        thisFixed = fixedsSoFar.shift()
        theLength = thisFixed.length
        var startOfText = textSoFar.substring(0,theLength)
        if (startOfText != thisFixed) {
            return -1
        } else {
            textSoFar = textSoFar.substring(theLength)
        }
        
        // now we have left v F v F v so peel off vars one by one
        while (fixedsSoFar.length > 0) {
            thisFixed = fixedsSoFar.shift()
            var thePos = textSoFar.indexOf(thisFixed)
            if (thePos == -1) {
                return -1
            } else {
                varsSoFar.push(textSoFar.substring(0,thePos))
                textSoFar = textSoFar.substring(thePos+thisFixed.length)
            }
        }
        
        varsSoFar.push(textSoFar)
    }
    
    return varsSoFar
}

// ------------- this routine reads in the syntaxes of the commands -----------

function pa_fetchSyntax() {
    var soFar = []
    
    for (var i=0; i<gLanguageSteps.length; i++) {
        var thisStepLeft = gLanguageSteps[i].parseText
        var thisStepFixeds = [gLanguageSteps[i].theKey]
        
        while (thisStepLeft.indexOf('%') != -1) {
            var thePos = thisStepLeft.indexOf('%')
            thisStepFixeds.push(thisStepLeft.slice(0,thePos))
            thisStepLeft = thisStepLeft.slice(thePos+2)
        }
        thisStepFixeds.push(thisStepLeft)
        
        soFar.push(thisStepFixeds)
    }
    return soFar
}

// ---------------------- compilation ---------------------

// the testing to see which template fits a line of script
// is done by checking each step type from the last to the first
// therefore two steptypes which might both fit have to be listed
// with the simpler one first and the more complicated one later
// for instance:
//        "set %1 to %2"
//        "set %1 to the length of %2"
// so that it checks for the 'length' part first and correctly
// identifies the script line as fitting the second template
// even though it fits the first template too.

function pa_compile(scriptAsBlocks) {
    var definesSoFar = []
    var tablesSoFar = []
    var textBlocksSoFar = []
    var stepsSoFar = [{
                        stepTokens: [3,"main",'find me'],
                        nestLevel: 0,
                        stepNumber: "root:0",
                        stepText: "runtime calling main"}]

    for (var i=0; i<scriptAsBlocks.length; i++) {

        switch (scriptAsBlocks[i].blockType) {

            case "proc":
                var procName = scriptAsBlocks[i].blockName
                var startLine = scriptAsBlocks[i].startLine
                var stepsAsParams = scriptAsBlocks[i].linesAsParams.slice()
                var stepsAsText = scriptAsBlocks[i].linesAsText.slice()
                var currentNestLevel = 0
                                
                for (var j=0; j<stepsAsParams.length; j++) {
                    var theseParams = stepsAsParams[j]
                    var theseTokens = pa_paramsToTokens(theseParams)
                    
                    if ((theseTokens[0] != 7) && (theseTokens[0] != 8)) {
                        currentNestLevel = calcNestLevelInto(currentNestLevel,theseParams[0])
                        stepsSoFar.push({
                            stepTokens: theseTokens,
                            nestLevel: currentNestLevel,
                            scriptLineNum: startLine+1+j,
                            stepText: stepsAsText[j]})
                        currentNestLevel = calcNestLevelOutof(currentNestLevel,theseParams[0])
                    }
                }
                break
                
            case "table":
                var tableName = scriptAsBlocks[i].blockName
                var tableOrder = "given"
                if (tableName.indexOf(" in random order") != -1) {
                    tableOrder = "random"
                    tableName = tableName.substring(0,tableName.length-16)
                }
                
                var tableContent = scriptAsBlocks[i].rowsAsArray
                tablesSoFar.push({
                    name:tableName,
                    order:tableOrder,
                    content:tableContent,unused:[]})
                break
                
            case "textblock":
                textBlocksSoFar.push({
                    name:scriptAsBlocks[i].blockName,
                    theKey:scriptAsBlocks[i].blockName,
                    content:scriptAsBlocks[i].content.join(gReturn)})
                break
                
            case "defineline":
                definesSoFar.push({
                    theKey:scriptAsBlocks[i].blockName,
                    content:scriptAsBlocks[i].content})
                break
            
            case "blankline":
            case "commentline":
            case "comment":
                break
        }
    }
    
    return {textBlocks: textBlocksSoFar,
            tables: tablesSoFar,
            steps: stepsSoFar,
            defines: definesSoFar}
}

function calcNestLevelInto(raw, theToken) {
    var temp = util_lookInObjList(gStepTypeNesting, theToken, {unstack:"", stack:""})
    if (temp.unstack == "") return raw
    return raw - 1
}

function calcNestLevelOutof(raw, theToken) {
    var temp = util_lookInObjList(gStepTypeNesting, theToken, {unstack:"", stack:""})
    if (temp.stack == "") return raw
    return raw + 1
}

// numeric entries in the schema refer to parameter numbers
// string entries in the scheme should be transferred untouched
function pa_paramsToTokens(theParams) {
    // first find the appropriate step template
    var stepType = theParams[0]
    var theTemplate = util_lookInObjList(gLanguageSteps, stepType).tokenParams
    if (theTemplate) {

        // now replace parameters with their values
        var soFar = theTemplate.slice()
        for (var i=1; i<theTemplate.length; i++) {
            if (typeof(theTemplate[i]) == 'number') soFar[i] = theParams[soFar[i]]
        }
        return soFar
    
    } else {
    
        alert ("Can't find template for command '"+stepType+"'.")
    }
}

// ------------------------ unparsing ----------------------
    
function pa_unparse(blocks) {
    soFar = ""
    var lastBlockType = 'single line'
    
    for (var i=0; i<blocks.length; i++) {
        switch (blocks[i].blockType) {

            case "blankline":
                var thisLine = ""
                                
                lastBlockType = 'single line'
                break
                
            case "commentline":
                var thisLine = "-- " + blocks[i].content
                
                lastBlockType = 'single line'
                break
                
            case "defineline":
                var thisLine = "define " + blocks[i].blockName +" "+ blocks[i].content
                
                lastBlockType = 'single line'
                break
                
            case "commentblock":
                var thisLine = (lastBlockType == 'block')?gReturn:""
            
                thisLine += "comment " + blocks[i].blockName + gReturn
                thisLine += blocks[i].content.join(gReturn) + gReturn
                thisLine += "end comment"
                
                lastBlockType = 'block'
                break
                
            case "textblock":
                var thisLine = (lastBlockType == 'block')?gReturn:""

                thisLine += "textblock " + blocks[i].blockName + gReturn
                thisLine += blocks[i].content.join(gReturn) + gReturn
                thisLine += "end textblock"
                
                lastBlockType = 'block'
                break
                
            case "table":
                var thisLine = (lastBlockType == 'block')?gReturn:""

                thisLine += "table " + blocks[i].blockName
                if (blocks[i].tableOrder == 'random') {
                    thisLine += " in random order"
                }
                thisLine += gReturn
                
                for (var j=0; j<blocks[i].content.length; j++) {
                    thisLine += gIndentSpaces+gIndentSpaces+blocks[i].content[j].join(",") + gReturn
                }
                thisLine += "end table"
                
                lastBlockType = 'block'
                break
                
            case "proc":
                var thisLine = (lastBlockType == 'block')?gReturn:""

                var indentSoFar = ""
                for (var j=0; j<blocks[i].linesAsParams.length; j++) {
                    var theParams = blocks[i].linesAsParams[j]
                    
                    var theseIndents = util_lookInObjList(gStepTypeIndents, theParams[0], {outdent: 0, indent: ""})
                    
                    indentSoFar = indentSoFar.substr(0,indentSoFar.length-theseIndents.outdent)
                    thisLine += indentSoFar + figureOutTextForParams(theParams) + gReturn
                    indentSoFar += theseIndents.indent
                }
                
                thisLine = thisLine.substr(0,thisLine.length-1)
                lastBlockType = 'block'
                break
                
            default:
                alert("Don't know how to handle block of type "+blocks[i].blockType+".")
        }
        soFar += thisLine + gReturn
    }
    
    return soFar.substr(0,soFar.length-1)
}

function figureOutTextForParams(theParams) {
    var theTemplate = util_lookInObjList(gLanguageSteps, theParams[0])
    
    var soFar = theTemplate.parseText
    for (var i=1; i<theParams.length; i++) {
        var theMarker = "%"+i
        soFar = util_replaceAll(soFar, theMarker, theParams[i])
    }
    
    return soFar
}
