// the testing to see which template fits a line of script
// is done by checking each step type from the last to the first
// therefore two steptypes which might both fit have to be listed
// with the simpler one first and the more complicated one later
// for instance:
//        "set %1 to %2"
//        "set %1 to the length of %2"
// so that it checks for the 'length' part first and correctly
// identifies the script line as fitting the second template
// even though it fits the first template too.

gLanguageSteps = [
	{theKey: "proc",
		parseText: "proc %1",
		tokenParams: [1,1],
		exampleText: "proc setUpDisplay"},
	{theKey: "procEnd",
		parseText: "end proc",
		tokenParams: [2],
		exampleText: "end proc"},
	{theKey: "call",
		parseText: "call %1",
		tokenParams: [3,1,'find me'],
		exampleText: "call procToCall"},
	{theKey: "halt",
		parseText: "halt",
		tokenParams: [5],
		exampleText: "halt"},
	{theKey: "blankLine",
		parseText: "",
		tokenParams: [7],
		exampleText: ""},
	{theKey: "commentLine",
		parseText: "-- %1",
		tokenParams: [8],
		exampleText: "-- something informative goes here"},
	{theKey: "declare",
		parseText: "declare %1",
		tokenParams: [11,1],
		exampleText: "declare $varCondition"},
	{theKey: "log",
		parseText: "log %1",
		tokenParams: [13,"",1],
        exampleText: "log $dateStamp"},
    {theKey: "logln",
      parseText: "logln %1",
      tokenParams: [13,"ln",1],
      exampleText: "logln $dateStamp"},
    {theKey: "log array",
      parseText: "log array %1",
      tokenParams: [13,"array",1],
      exampleText: "log array $varSortedResponses"},
	{theKey: "logToConsole",
		parseText: "log to console %1",
		tokenParams: [14,"",1],
		exampleText: "log to console $lastKey"},
	{theKey: "background",
		parseText: "background (%1,%2,%3)",
		tokenParams: [20,1,2,3],
		exampleText: "background (50,150,200)"},
	{theKey: "backgroundGrey",
		parseText: "background grey",
		tokenParams: [20,"127","127","127"],
		exampleText: "background grey"},
	{theKey: "backgroundWhite",
		parseText: "background white",
		tokenParams: [20,"255","255","255"],
		exampleText: "background white"},
	{theKey: "backgroundBlack",
		parseText: "background black",
		tokenParams: [20,"0","0","0"],
		exampleText: "background black"},
	{theKey: "displayTextFromTextblock",
		parseText: "display text from textblock %1",
		tokenParams: [21,1],
        exampleText: "display text from textblock thankyou"},
    {theKey: "displayTextFromVariable",
        parseText: "display text from variable %1",
        tokenParams: [22,1],
        exampleText: "display text from variable thankyou"},
    {theKey: "displaySurvey",
        parseText: "display survey from table %1",
        tokenParams: [23,1],
        exampleText: "display survey from table surveyquestions"},
    {theKey: "cursorOther",
        parseText: "cursor %1",
        tokenParams: [24,1],
        exampleText: "cursor pointer"},
    {theKey: "cursorOff",
        parseText: "cursor off",
        tokenParams: [24,'none'],
        exampleText: "cursor off"},
    {theKey: "cursorOn",
        parseText: "cursor on",
        tokenParams: [24,'default'],
        exampleText: "cursor on"},
	{theKey: "setTo",
		parseText: "set %1 to %2",
		tokenParams: [33,"to",1,2],
		exampleText: "set $varTime to $timeStamp"},
	{theKey: "setByAppending",
		parseText: "set %1 by appending %2",
		tokenParams: [33,"appending",1,2],
		exampleText: "set $varCounter by appending X"},
	{theKey: "setByReplacing",
		parseText: "set %1 by replacing %2 with %3",
		tokenParams: [33,"replacing",1,2,3],
		exampleText: "set $varCounter by replacing X with Y"},
	{theKey: "setToLength",
		parseText: "set %1 to the length of %2",
		tokenParams: [33,"length",1,2],
		exampleText: "set $vaKeyCount to the length of $varKeysSoFar"},
	{theKey: "setToCharacters",
		parseText: "set %1 to characters %2 to %3 of %4",
		tokenParams: [33,"characters",1,2,3,4],
		exampleText: "set $varSide to characters 2 to 2 of ILM"},
	{theKey: "setArrayTo",
        parseText: "set array %1 to %2",
        tokenParams: [34,"to",1,2],
        exampleText: "set array $varTimeOptions to |early|on time|late"},
    {theKey: "extractFirstItemOfArrayTo",
        parseText: "extract first item of array %1 to %2",
        tokenParams: [35,"extractfirst",1,2],
        exampleText: "extract first item of array $varListOfAnswers to $varThisAnswer"},
    {theKey: "setArrayByAppending",
        parseText: "set array %1 by appending %2",
        tokenParams: [34,"appending",1,2],
        exampleText: "set array $varKeysSoFar by appending $lastKey"},
	{theKey: "setArrayToLength",
		parseText: "set %1 to the length of array %2",
		tokenParams: [34,"length",1,2],
          exampleText: "set $varKeyCount to the length of array $varKeysSoFar"},
    {theKey: "setToItemOfArray",
          parseText: "set %1 to item %2 of array %3",
        tokenParams: [34,"toitemofarray",1,2,3],
        exampleText: "set $varSide to item 2 of L|R|C"},
    {theKey: "setItemOfArrayTo",
          parseText: "set item %1 of array %2 to %3",
        tokenParams: [34,"itemofarrayto",1,2,3],
        exampleText: "set item 2 of array $varSide to Right"},
	{theKey: "ifIs",
		parseText: "if %1 is %2",
		tokenParams: [41,"is",1,2,'find me'],
		exampleText: "if $varSide is left"+gReturn+"else"+gReturn+"end if"},
	{theKey: "ifIsNot",
		parseText: "if %1 is not %2",
		tokenParams: [41,"is not",1,2,'find me'],
		exampleText: "if $varSide is not left"+gReturn+"else"+gReturn+"end if"},
	{theKey: "ifIsIn",
		parseText: "if %1 is in %2",
		tokenParams: [41,"is in",1,2,'find me'],
		exampleText: "if apple is in $varFruitList"+gReturn+"else"+gReturn+"end if"},
	{theKey: "ifIsNotIn",
		parseText: "if %1 is not in %2",
		tokenParams: [41,"is not in",1,2,'find me'],
		exampleText: "if apple is not in $varFruitList"+gReturn+"else"+gReturn+"end if"},
	{theKey: "ifElse",
		parseText: "else",
		tokenParams: [42,"always",'x','y','find me'],
		exampleText: "else"},
	{theKey: "ifElseIs",
		parseText: "else if %1 is %2",
		tokenParams: [42,"is",1,2,'find me'],
		exampleText: "else if $varSide is left"},
	{theKey: "ifElseIsNot",
		parseText: "else if %1 is not %2",
		tokenParams: [42,"is not",1,2,'find me'],
		exampleText: "else if $varSide is not left"},
	{theKey: "ifElseIsIn",
		parseText: "else if %1 is in %2",
		tokenParams: [42,"is in",1,2,'find me'],
		exampleText: "else if apple is in $varFruitList"},
	{theKey: "ifElseIsNotIn",
		parseText: "else if %1 is not in %2",
		tokenParams: [42,"is not in",1,2,'find me'],
		exampleText: "else if apple is not in $varFruitList"},
	{theKey: "ifEnd",
		parseText: "end if",
		tokenParams: [49],
		exampleText: "end if"},
	{theKey: "whileIs",
		parseText: "while %1 is %2",
		tokenParams: [51,"is",1,2],
		exampleText: "while $varSide is left"+gReturn+"end while"},
	{theKey: "whileIsNot",
		parseText: "while %1 is not %2",
		tokenParams: [51,"is not",1,2],
		exampleText: "while $varSide is not left"+gReturn+"end while"},
	{theKey: "whileIsIn",
		parseText: "while %1 is in %2",
		tokenParams: [51,"is in",1,2],
		exampleText: "while apple is in $varFruitList"+gReturn+"end while"},
	{theKey: "whileIsNotIn",
		parseText: "while %1 is not in %2",
		tokenParams: [51,"is not in",1,2],
		exampleText: "while apple is not in $varFruitList"+gReturn+"end while"},
	{theKey: "whileEnd",
		parseText: "end while",
		tokenParams: [53,'find me'],
		exampleText: "end while"},
	{theKey: "repeatTimes",
		parseText: "repeat %1 times",
		tokenParams: [56,1],
		exampleText: "repeat 10 times"+gReturn+"end repeat"},
	{theKey: "repeatTable",
		parseText: "repeat for each row in table %1",
		tokenParams: [57,1,'find me'],
		exampleText: "repeat for each row in table stims"+gReturn+"end repeat"},
	{theKey: "repeatEnd",
		parseText: "end repeat",
		tokenParams: [58,'find me'],
        exampleText: "end repeat"},
    {theKey: "resetTable",
        parseText: "reset table %1",
        tokenParams: [61,'find me',1],
        exampleText: "reset table stimuli"},
    {theKey: "readFromTable",
        parseText: "read %1 from table %2",
        tokenParams: [62,2,'find me',1],
        exampleText: "read $varFilename,$varCorrectClick from table stimuli"},
    {theKey: "moveCell",
		parseText: "move cell %1 to (%2,%3)",
		tokenParams: [70,1,2,3],
		exampleText: "move cell A to (0,0)"},
	{theKey: "moveCellToCentre",
		parseText: "move cell %1 to centre",
		tokenParams: [70,1,"0","0"],
		exampleText: "move cell A to centre"},
	{theKey: "showCell",
		parseText: "show cell %1",
		tokenParams: [71,1,"show"],
		exampleText: "show cell A"},
	{theKey: "hideCell",
		parseText: "hide cell %1",
		tokenParams: [71,1,"hide"],
		exampleText: "hide cell A"},
	{theKey: "loadCellWithImage",
		parseText: "load cell %1 with image %2",
		tokenParams: [72,1,2,""],
		exampleText: "load cell C with image image.png"},
	{theKey: "loadCellWithImageFromFolder",
		parseText: "load cell %1 with image %2 from folder %3",
		tokenParams: [72,1,2,3],
		exampleText: "load cell A with image happyface12.png from folder faces"},
	{theKey: "loadCellWithMovie",
		parseText: "load cell %1 with movie %2",
		tokenParams: [73,1,2,""],
		exampleText: "load cell A with movie plan9.mov"},
	{theKey: "loadCellWithMovieFromFolder",
		parseText: "load cell %1 with movie %2 from folder %3",
		tokenParams: [73,1,2,3],
		exampleText: "load cell A with movie plan9.mov from folder monochrome"},
	{theKey: "playMovie",
		parseText: "play movie in cell %1",
		tokenParams: [74,1,"wait"],
		exampleText: "play movie in cell A"},
	{theKey: "playMovieNoWait",
		parseText: "play movie in cell %1 without waiting",
		tokenParams: [74,1,"nowait"],
		exampleText: "play movie in cell A without waiting"},
	{theKey: "pauseMovie",
		parseText: "pause movie in cell %1",
		tokenParams: [75,1],
		exampleText: "pause movie in cell A"},
	{theKey: "loadCellWithText",
		parseText: "load cell %1 with text %2 at length %3",
		tokenParams: [76,1,2,3],
		exampleText: "load cell C with text dummytext at length 20"},
	{theKey: "loadCellWithSound",
		parseText: "load cell %1 with sound %2",
		tokenParams: [77,1,2,""],
		exampleText: "load cell A with sound beep.mp3"},
	{theKey: "loadCellWithSoundFromFolder",
		parseText: "load cell %1 with sound %2 from folder %3",
		tokenParams: [77,1,2,3],
		exampleText: "load cell A with sound tring.mp4 from folder artificial"},
	{theKey: "playSound",
		parseText: "play sound in cell %1",
		tokenParams: [78,1,"wait"],
		exampleText: "play sound in cell A"},
	{theKey: "playSoundNoWait",
		parseText: "play sound in cell %1 without waiting",
		tokenParams: [78,1,"nowait"],
		exampleText: "play sound in cell A without waiting"},
	{theKey: "pauseSound",
		parseText: "pause sound in cell %1",
		tokenParams: [79,1],
		exampleText: "pause sound in cell A"},
	{theKey: "loadCellWithSlider",
		parseText: "load cell %1 with slider from %2 to %3",
		tokenParams: [81,1,"horizontal",2,3,"none"],
		exampleText: "load cell R with slider from 0 to 99"},
	{theKey: "loadCellWithSliderStep",
		parseText: "load cell %1 with slider from %2 to %3 step %4",
		tokenParams: [81,1,"horizontal",2,3,4],
		exampleText: "load cell R with slider from 0 to 99 step 10"},
	{theKey: "loadCellWithHSlider",
		parseText: "load cell %1 with horizontal slider from %2 to %3",
		tokenParams: [81,1,"horizontal",2,3,"none"],
		exampleText: "load cell R with horizontal slider from 0 to 99"},
	{theKey: "loadCellWithHSliderStep",
		parseText: "load cell %1 with horizontal slider from %2 to %3 step %4",
		tokenParams: [81,1,"horizontal",2,3,4],
		exampleText: "load cell R with horizontal slider from 0 to 99 step 10"},
	{theKey: "loadCellWithHSlider",
		parseText: "load cell %1 with vertical slider from %2 to %3",
		tokenParams: [81,1,"vertical",2,3,"none"],
		exampleText: "load cell R with vertical slider from 0 to 99"},
	{theKey: "loadCellWithHSliderStep",
		parseText: "load cell %1 with vertical slider from %2 to %3 step %4",
		tokenParams: [81,1,"vertical",2,3,4],
		exampleText: "load cell R with vertical slider from 0 to 99 step 10"},
	{theKey: "setSliderCellTo",
		parseText: "set slider cell %1 to %2",
		tokenParams: [82,1,2],
		exampleText: "set slider cell R to 24"},
	{theKey: "setTextCellTo",
		parseText: "set text cell %1 to %2",
		tokenParams: [83,1,2],
		exampleText: "set text cell R to Finished."},
	{theKey: "waitForAClick",
		parseText: "wait for a click",
		tokenParams: [91,"ABCDEFGHIJKLMNOPQRSTUVWXYZ","none","none"],
		exampleText: "wait for a click"},
	{theKey: "waitForAClickIn",
		parseText: "wait for a click in %1",
		tokenParams: [91,1,"none","none"],
		exampleText: "wait for a click in ABC"},
	{theKey: "waitForAKey",
		parseText: "wait for a key",
		tokenParams: [91,"none","1234567890-=qwertyuiop[]asdfghjkl;'`zxcvbnm,./ +","none"],
		exampleText: "wait for a key"},
	{theKey: "waitForAKeyIn",
		parseText: "wait for a key in %1",
		tokenParams: [91,"none",1,"none"],
		exampleText: "wait for a key in YN"},
	{theKey: "waitForTime",
		parseText: "wait for %1 second",
		tokenParams: [91,"none","none",1],
		exampleText: "wait for 5 second"},
	{theKey: "waitForTimes",
		parseText: "wait for %1 seconds",
		tokenParams: [91,"none","none",1],
      exampleText: "wait for 5 seconds"},
    {theKey: "waitForAClickInTime",
        parseText: "wait for a click in %1 or for %2 second",
        tokenParams: [91,1,"none",2],
        exampleText: "wait for a click in ABC or for 1 second"},
    {theKey: "waitForTimeAClickIn",
        parseText: "wait for %1 seconds or a click in %2",
        tokenParams: [91,1,"none",2],
        exampleText: "wait for a click in ABC or for 1 seconds"},
    {theKey: "waitForAKeyTime",
        parseText: "wait for a key or for %1 seconds",
        tokenParams: [91,"none","1234567890-=qwertyuiop[]asdfghjkl;'`zxcvbnm,./ +",1],
        exampleText: "wait for a key or for 1 seconds"},
    {theKey: "waitForAKeyInTime",
        parseText: "wait for a key in %1 or for %2 seconds",
        tokenParams: [91,"none",1,2],
        exampleText: "wait for a key in YN or for 1 seconds"},
    {theKey: "waitForTimeAKeyIn",
        parseText: "wait for %1 seconds or for a key in %2",
        tokenParams: [91,"none",2,1],
        exampleText: "wait for a key in YN or for 1 seconds"},
    {theKey: "waitForAClickKeyTime",
        parseText: "wait for a click in %1 or a key in %2 or for %3 seconds",
        tokenParams: [91,1,2,3],
        exampleText: "wait for a click in ABCDE or a key in abcde or for 1 seconds"},
    {theKey: "waitForAClickInTimes",
        parseText: "wait for a click in %1 or for %2 seconds",
        tokenParams: [91,1,"none",2],
        exampleText: "wait for a click in ABC or for 22 seconds"},
    {theKey: "waitForTimesAClickIn",
        parseText: "wait for %1 seconds or a click in %2",
        tokenParams: [91,1,"none",2],
        exampleText: "wait for a click in ABC or for 22 seconds"},
    {theKey: "waitForAKeyTimes",
        parseText: "wait for a key or for %1 seconds",
        tokenParams: [91,"none","1234567890-=qwertyuiop[]asdfghjkl;'`zxcvbnm,./ +",1],
        exampleText: "wait for a key or for 3.6 seconds"},
    {theKey: "waitForAKeyInTimes",
        parseText: "wait for a key in %1 or for %2 seconds",
        tokenParams: [91,"none",1,2],
        exampleText: "wait for a key in YN or for 3.6 seconds"},
    {theKey: "waitForTimesAKeyIn",
        parseText: "wait for %1 seconds or for a key in %2",
        tokenParams: [91,"none",2,1],
        exampleText: "wait for a key in YN or for 3.6 seconds"},
    {theKey: "waitForAClickKeyTimes",
        parseText: "wait for a click in %1 or a key in %2 or for %3 seconds",
        tokenParams: [91,1,2,3],
        exampleText: "wait for a click in ABCDE or a key in abcde or for 60 seconds"},
	{theKey: "startTimer",
		parseText: "start timer",
		tokenParams: [96],
		exampleText: "start timer"},
	{theKey: "countDownFroms",
		parseText: "count down from %1 seconds",
		tokenParams: [97,1],
		exampleText: "count down from 10 seconds"},
	{theKey: "countDownFrom",
		parseText: "count down from %1 second",
		tokenParams: [97,1],
		exampleText: "count down from 1 second"}
]

// ---------------------- formatting information for tidying up -------------

    gStepTypeNesting = [
        {theKey: "proc",          unstack:"",      stack:"proc"},
        {theKey: "procEnd",       unstack:"proc",  stack:""},
        {theKey: "ifIs",          unstack:"",      stack:"if"},
        {theKey: "ifIsNot",       unstack:"",      stack:"if"},
        {theKey: "ifIsIn",        unstack:"",      stack:"if"},
        {theKey: "ifIsNotIn",     unstack:"",      stack:"if"},
        {theKey: "ifElse",        unstack:"if",    stack:"if"},
        {theKey: "ifElseIs",      unstack:"if",    stack:"if"},
        {theKey: "ifElseIsNot",   unstack:"if",    stack:"if"},
        {theKey: "ifElseIsIn",    unstack:"if",    stack:"if"},
        {theKey: "ifElseIsNotIn", unstack:"if",    stack:"if"},
        {theKey: "ifEnd",         unstack:"if",    stack:""},
        {theKey: "whileIs",       unstack:"",      stack:"while"},
        {theKey: "whileIsNot",    unstack:"",      stack:"while"},
        {theKey: "whileIsIn",     unstack:"",      stack:"while"},
        {theKey: "whileIsNotIn",  unstack:"",      stack:"while"},
        {theKey: "whileEnd",      unstack:"while", stack:""},
        {theKey: "repeatTimes",   unstack:"",      stack:"repeat"},
        {theKey: "repeatTable",   unstack:"",      stack:"repeat"},
        {theKey: "repeatEnd",     unstack:"repeat",stack:""}
    ]

    gStepTypeIndents = [
        {theKey:"proc", outdent:0, indent:gIndentSpaces+gIndentSpaces},
        {theKey:"procEnd", outdent:gIndentSpacesLength+gIndentSpacesLength, indent:""},
        {theKey:"ifIs", outdent:0, indent:gIndentSpaces+gIndentSpaces},
        {theKey:"ifIsNot", outdent:0, indent:gIndentSpaces+gIndentSpaces},
        {theKey:"ifIsIn", outdent:0, indent:gIndentSpaces+gIndentSpaces},
        {theKey:"ifIsNotIn", outdent:0, indent:gIndentSpaces+gIndentSpaces},
        {theKey:"ifElse", outdent:gIndentSpacesLength, indent:gIndentSpaces},
        {theKey:"ifElseIs", outdent:gIndentSpacesLength, indent:gIndentSpaces},
        {theKey:"ifElseIsNot", outdent:gIndentSpacesLength, indent:gIndentSpaces},
        {theKey:"ifElseIsIn", outdent:gIndentSpacesLength, indent:gIndentSpaces},
        {theKey:"ifElseIsNotIn", outdent:gIndentSpacesLength, indent:gIndentSpaces},
        {theKey:"ifEnd", outdent:gIndentSpacesLength+gIndentSpacesLength, indent:""},
        {theKey:"whileIs", outdent:0, indent:gIndentSpaces},
        {theKey:"whileIsNot", outdent:0, indent:gIndentSpaces},
        {theKey:"whileIsIn", outdent:0, indent:gIndentSpaces},
        {theKey:"whileIsNotIn", outdent:0, indent:gIndentSpaces},
        {theKey:"whileEnd", outdent:gIndentSpacesLength, indent:""},
        {theKey:"repeatTimes", outdent:0, indent:gIndentSpaces},
        {theKey:"repeatTable", outdent:0, indent:gIndentSpaces},
        {theKey:"repeatEnd", outdent:gIndentSpacesLength, indent:""}
    ]
